﻿using System;
using System.Linq;
using System.Threading;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.SharePoint.ApplicationRuntime;
using Microsoft.SharePoint.Utilities;
using Microsoft.SharePoint.WebControls;
using RH.Warmup.Helper;
using RH.Warmup.TimerJob;
using RH.Warmup.Update;
using RH.Warmup.Web.UI;
using WarmupSharePoint;
using WarmupSharePoint.Helper;

namespace RH.Warmup
{
	public partial class ConfigureWarmupJob : LayoutsPageBase
	{
		private Settings _repeaterSettings;
		private bool _error;

		public ConfigureWarmupJob()
		{
			_repeaterSettings = new Settings();
		}

		protected override void OnInit(EventArgs e)
		{
			base.OnInit(e);

			// need to ensure control here, so the changed events within WarmupSettingsItemTemplate will work
			EnsureChildControls();
		}

		protected override void CreateChildControls()
		{
			try
			{
				base.CreateChildControls();

				// get settings from Farm
				_repeaterSettings = LoadSettings();

				// add all other WebApplications for them being selectable
				var service = SPFarm.Local.Services.GetValue<SPWebService>("");
				foreach (SPWebApplication webApplication in service.WebApplications)
				{
					Setting setting = _repeaterSettings.SingleOrDefault(s => s.WebApplicationId == webApplication.Id);
					if (setting != null)
					{
						// update name for GUI
						setting.Name = webApplication.Name;
						GetTimerJobId(webApplication, setting);
					}
					else
					{
						// add to Repeater, so currently not selected WebApplications are available for configurations
						_repeaterSettings.Add(new Setting
						{
							WebApplicationId = webApplication.Id,
							Name = webApplication.Name
						});
					}
				}

				// configure Repeater to show the WebApplications
				WarmupSettingsRepeater.ItemTemplate = new WarmupSettingsItemTemplate { RowCount = _repeaterSettings.Count };
				WarmupSettingsRepeater.AlternatingItemTemplate = new WarmupSettingsItemTemplate { RowCount = _repeaterSettings.Count, Alternating = true };
				WarmupSettingsRepeater.DataSource = _repeaterSettings;

				var updates = new Updates("RH.Warmup", null);
				var newestVersion = updates.GetLatestVersion();

				if (!string.IsNullOrEmpty(newestVersion.ErrorText))
				{
					ProduktVersion.Text = newestVersion.ErrorText;
				}
				else if (newestVersion.Version > ProductVersion.Version)
				{
					ProduktVersion.Text = LocalizationHelper.GetString("ApplicationPage_Version_UpdateAvailable");
				}
				else
				{
					ProduktVersion.Text = LocalizationHelper.GetString("ApplicationPage_CurrentVersion") + ProductVersion.Version;
				}
			}
			catch (Exception ex)
			{
				LabelErrorMessage.Text = ex.ToString();
			}
		}

		protected override void OnPreRender(EventArgs e)
		{
			try
			{
				base.OnPreRender(e);

				WarmupSettingsRepeater.DataBind();
			}
			catch (Exception ex)
			{
				LabelErrorMessage.Text = ex.ToString();
			}
		}

		protected void BtnSubmit_Click(object sender, EventArgs e)
		{
			try
			{
				UpdateSettingsInFarm();

				if (!_error)
					SPSecurity.RunWithElevatedPrivileges(ConfigureTimerJob);

				if (!_error)
					SPUtility.Redirect("/monitoring.aspx", SPRedirectFlags.Default, Context);
			}
			catch (ThreadAbortException)
			{
				// occures if redirected
			}
			catch (Exception ex)
			{
				LabelErrorMessage.Text = ex.ToString();
			}
		}

		/// <summary>
		/// get TimerJobId for existing jobs
		/// </summary>
		/// <param name="webApplication"></param>
		/// <param name="setting"></param>
		private void GetTimerJobId(SPWebApplication webApplication, Setting setting)
		{
			foreach (SPJobDefinition jobDefinition in SPFarm.Local.TimerService.JobDefinitions)
			{
				if (jobDefinition.DisplayName == GetJobName(webApplication))
				{
					setting.TimerJobId = jobDefinition.Id;
					return;
				}
			}
		}

		private void ConfigureTimerJob()
		{
			var jobDefinitions = SPFarm.Local.TimerService.JobDefinitions;

			// remove existing jobs and readd it
			foreach (SPJobDefinition jobDefinition in jobDefinitions.Where(j => j.Name.StartsWith(WarmupJob.JobTitle)))
			{
				jobDefinitions.Remove(jobDefinition.Id);
			}

			// create TimerJobs for each selected WebApplication
			var service = SPFarm.Local.Services.GetValue<SPWebService>("");
			var was = from webApplication in service.WebApplications
						 let setting = _repeaterSettings.Find(s => s.Activated && s.WebApplicationId == webApplication.Id)
						 where _repeaterSettings.Find(s => s.Activated && s.WebApplicationId == webApplication.Id) != null
						 select new { webApplication, setting };
			foreach (var wa in was)
			{
				var warmupJob = new WarmupJob(GetJobName(wa.webApplication), wa.webApplication);
				var startTime = wa.setting.Time;

				if (wa.setting.Schedule == 24)
				{
					var endTime = startTime.AddMinutes(15);
					warmupJob.Schedule = new SPDailySchedule { BeginHour = startTime.Hour, BeginMinute = startTime.Minute, EndHour = endTime.Hour, EndMinute = endTime.Minute };
				}
				else
				{
					warmupJob.Schedule = new SPHourlySchedule { BeginMinute = startTime.Minute, EndMinute = startTime.Minute + 1 };
				};
				// add custom settings to job, so it will know what to do
				warmupJob.Properties.Add(WarmupJob.PropertyName, JsonHelper.JsonSerializer(_repeaterSettings.Find(s => s.WebApplicationId == wa.webApplication.Id)));
				jobDefinitions.Add(warmupJob);
			}
		}

		private string GetJobName(SPWebApplication webApplication)
		{
			var jobName = string.Format("{0} - '{1}'", WarmupJob.JobTitle, webApplication.Name);
			return jobName;
		}

		/// <summary>
		/// Get Settings from Farm PropertyBag
		/// </summary>
		/// <returns></returns>
		internal static Settings LoadSettings()
		{
			object farmProperty = null;
			SPSecurity.RunWithElevatedPrivileges(() =>
			{
				farmProperty = SPFarm.Local.Properties[Constants.PropertyBagKey];
			});
			if (farmProperty == null)
				return new Settings();

			var settings = JsonHelper.JsonDeserialize<Settings>(farmProperty.ToString());
			// set activated property on loaded settings
			settings.ForEach(s => s.Activated = true);
			return settings;
		}

		/// <summary>
		/// Store Settings in Farm PropertyBag
		/// </summary>
		internal void UpdateSettingsInFarm()
		{
			UpdateSettingsWithCurrentValues();

			SPSecurity.RunWithElevatedPrivileges(() =>
			{
				// need to store activated WebApplications only
				var activatedWebApplications = new Settings();
				activatedWebApplications.AddRange(_repeaterSettings.Where(s => s.Activated).ToList());
				string jsonString = JsonHelper.JsonSerializer(activatedWebApplications);

				if (!SPFarm.Local.Properties.ContainsKey(Constants.PropertyBagKey))
				{
					SPFarm.Local.Properties.Add(Constants.PropertyBagKey, jsonString);
				}
				else
				{
					SPFarm.Local.Properties[Constants.PropertyBagKey] = jsonString;
				}
				SPFarm.Local.Update(true);
			});
		}

		/// <summary>
		/// Set current Values from Repeater in settings
		/// </summary>
		private void UpdateSettingsWithCurrentValues()
		{
			// iterate controls within RepeaterItems to fetch current values
			foreach (RepeaterItem item in WarmupSettingsRepeater.Controls)
			{
				try
				{
					if (item.ItemType != ListItemType.Item && item.ItemType != ListItemType.AlternatingItem)
						continue;

					var row = item.Controls[0] as HtmlTableRow;
					// get WebApplicationID from data attribute of the row, to find the settings which need to be updated
					var webApplicationId = new Guid(row.Attributes["data-webapplicationid"]);
					var setting = _repeaterSettings.Single(s => s.WebApplicationId == webApplicationId);

					foreach (HtmlTableCell cell in row.Controls)
					{
						var control = cell.Controls[0] as WebControl;
						switch (control.ID)
						{
							case "ActivateWarmup":
								setting.Activated = control.GetValue<CheckBox, bool>();
								break;
							case "WebApplicationName":
								break;
							case "WarmupTime":
								setting.Time = control.GetValue<TimePicker, DateTime>();
								setting.Schedule = int.Parse(((WebControl)cell.Controls[2]).GetValue<DropDownList, string>());
								break;
							case "SearchWarmup":
								setting.WarmupSearch = control.GetValue<CheckBox, bool>();
								break;
							case "SearchKeyword":
								setting.QueryKeyword = control.GetValue<TextBox, string>();
								break;
							case "Eventlog":
								setting.EventLog = control.GetValue<CheckBox, bool>();
								break;
						}
					}
				}
				catch (Exception ex)
				{
					_error = true;
					Trace.Write(ex.ToString());
					LabelErrorMessage.Text = ex.Message;
				}
			}
		}
	}
}
