﻿using System;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using Microsoft.SharePoint.Administration;
using WarmupSharePoint;
using WarmupSharePoint.Helper;

namespace RH.Warmup.TimerJob
{
	/// <summary>
	/// The TimerJob will be executed on each Server with Webfrontend role within the Farm
	/// </summary>
	[Guid("28E0F85D-8803-48B6-9798-A1BDA4B40AA1")] // This is a requirement for any class that derives directly or indirectly from SPPersistedObject.
	public class WarmupJob : SPJobDefinition
	{
		internal static string JobTitle = LocalizationHelper.GetString("TimerJob_Name");
		internal static string PropertyName = "RH.Warmup.WebApplication.Settings";

		public override string Description
		{
			get
			{
				return LocalizationHelper.GetString("TimerJob_Description");
			}
		}

		/// <summary>
		/// Default Constructor for serialization
		/// </summary>
		public WarmupJob()
		{
		}

		public WarmupJob(string jobName, SPService service, SPServer server, SPJobLockType targetType)
			: base(jobName, service, server, targetType)
		{
			Title = jobName;
		}

		// null as Server will execute the code on each WebFrontend
		public WarmupJob(string jobName, SPWebApplication webApplication)
			: base(jobName, webApplication, null, SPJobLockType.None)
		{
			Title = jobName;
		}

		public override void Execute(Guid targetInstanceId)
		{
			try
			{
				base.Execute(targetInstanceId);

				// custom settings from TimerJob
				var setting = JsonHelper.JsonDeserialize<Setting>(Convert.ToString(Properties[PropertyName]));

				var service = SPFarm.Local.Services.GetValue<SPWebService>(string.Empty);
				foreach (SPWebApplication webApplication in service.WebApplications.Where(wa => wa.Id == setting.WebApplicationId))
				{
					string[] iisUrls = SharePointHelper.GetAlternatUrlForWebApplication(webApplication);
					var errorMessages = new StringBuilder();
					var warmupWebsCount = 0;

					for (int index = 0; index < iisUrls.Length; index++)
					{
						string iisUrl = iisUrls[index];
						setting.WebApplicationUrl = iisUrl;

						var warmup = new WarmupSharePoint.Warmup(setting);
						warmup.ProgressChanged += delegate(object sender, ProgressChangeEventArgs args)
						{
							// calculate correct progress, in case of multiple iis urls
							var current = ((double)args.Progress + (index * args.Overall)) / ((iisUrls.Length + 1) * args.Overall);
							UpdateProgress((int)current);
						};
						warmup.ReadAllSites();

						errorMessages.Append(warmup.WarmupErrors);
						// overwrite with latest value, because it doesn't matter. all iis urls have the same amount of webs
						warmupWebsCount = warmup.WarmupWebsCount;
					}

					if (setting.EventLog)
					{
						var errors = errorMessages.ToString();
						var message = string.Format("Warmup results for '{0}'. Read {1} webs.", setting.WebApplicationUrl, warmupWebsCount);
						if (string.IsNullOrEmpty(errors))
						{
							Eventlog.Write(message, EventLogEntryType.Information, 0);
						}
						else
						{
							Eventlog.Write(string.Format("{0} Errors:{1}{2}", message, Environment.NewLine, errors), EventLogEntryType.Error, 1);
						}
					}
				}
			}
			catch (Exception ex)
			{
				Eventlog.Write(ex.ToString(), EventLogEntryType.Error, 1);
			}
		}
	}
}
