﻿using System;
using System.DirectoryServices;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Web.UI;
using System.Web.UI.WebControls;
using Microsoft.SharePoint.Administration;
using RH.Warmup.Helper;

namespace RH.Warmup.Web.UI
{
	/// <summary>
	/// custom TimeSlider with jQuery Plugin
	/// </summary>
	public class TimePicker : TextBox
	{
		private const string CacheKeyStartScript = "RH.Warmup.Web.UI.TimePickerStartScript";
		private const string CacheKeyAlreadyAdded = "RH.Warmup.Web.UI.TimePicker.AlreadyAdded";

		public TimePicker()
		{
			CssClass = "timePicker";
		}

		protected override void CreateChildControls()
		{
			base.CreateChildControls();
		}

		protected override void OnPreRender(EventArgs e)
		{
			base.OnPreRender(e);

			if (Value == DateTime.MinValue)
				Text = GetRecycleTimeFromIIS().ToString(CultureInfo.InvariantCulture);

			// remember each TimePicker and add one Startupscript for all pickers
			var addPicker = string.Format("      $rhWarmup.$('#{0}').rhWarmupTimePicker({{'hour':{1},'minute':{2},'startjobtext':'{3}','startedjobtext':'{4}'}});\r\n",
				ClientID, Value.Hour, Value.Minute, LocalizationHelper.GetString("ApplicationPage_Javascript_StartJob"), LocalizationHelper.GetString("ApplicationPage_Javascript_StartedJob"));
			var addPickers = Context.Items[CacheKeyStartScript] as string ?? string.Empty;
			addPickers += addPicker;
			Context.UpdateOrInsertCache(CacheKeyStartScript, addPickers);

			var alreadyAdded = Context.GetFromCache<int>(CacheKeyAlreadyAdded);
			alreadyAdded++;
			if (alreadyAdded < ControlCount)
			{
				// don't add the script now. wait until all control startup scripts have been collected
				Context.UpdateOrInsertCache(CacheKeyAlreadyAdded, alreadyAdded.ToString(CultureInfo.InvariantCulture));
			}
			else
			{
				//  add scripts only once
				//AddJavaScript("jquery-1.9.1.js", null, null);
				//AddJavaScript("RHWarmup.js", null, null);
				AddJavaScript("jquery-1.9.1.js", "rhWarmupJquery.js", null);
				AddJavaScript("RHWarmup.js", "rhWarmup.js", "rhWarmupJquery.js");

				// add one Startup Script for all picker instances
				var js = new StringBuilder(Environment.NewLine);
				js.Append("EnsureScriptFunc('rhWarmupJquery.js');\r\n");
				js.Append("ExecuteOrDelayUntilScriptLoaded(\r\n");
				js.Append("  function() {\r\n");
				js.Append("    EnsureScriptFunc('rhWarmup.js');\r\n");
				js.Append("    ExecuteOrDelayUntilScriptLoaded(\r\n");
				js.Append("      function() {\r\n");
				// wait for the document to be loaded, to be able to attach to controls
				js.Append("        $rhWarmup.$(document).ready(function() {\r\n");
				js.Append(addPickers);
				js.Append("        });\r\n");
				js.Append("      }\r\n");
				js.Append("    ,\"rhWarmup.js\")\r\n");
				js.Append("  }\r\n");
				js.Append(",\"rhWarmupJquery.js\")\r\n");
				ScriptManager.RegisterStartupScript(Page, typeof(TimePicker), "RH.TimePicker", js.ToString(), true);
			}
		}

		/// <summary>
		/// Get DateTIme Value from Text Property. Will be DateTime.MinValue if the value is not a valid DateTime
		/// </summary>
		public DateTime Value
		{
			get
			{
				DateTime time;
				if (DateTime.TryParse(Text, new DateTimeFormatInfo(), DateTimeStyles.None, out time))
					return time;

				return DateTime.MinValue;
			}
		}

		/// <summary>
		/// The TimePicker is used this often on the page
		/// </summary>
		public int ControlCount { get; set; }

		public Guid WebApplicationId { get; set; }

		private DateTime GetRecycleTimeFromIIS()
		{
			if (WebApplicationId == Guid.Empty)
				throw new InvalidOperationException("The WebApplicationID must not be null.");

			try
			{
				var service = SPFarm.Local.Services.GetValue<SPWebService>("");
				var webApplication = service.WebApplications.Single(w => w.Id == WebApplicationId);
				SPApplicationPool appPool = webApplication.ApplicationPool;

				var appPoolPath = "IIS://localhost/W3SVC/AppPools/" + appPool.Name;
				using (var applicationPool = new DirectoryEntry(appPoolPath))
				{
					PropertyValueCollection recycleTime = applicationPool.Properties["PeriodicRestartSchedule"];
					var value = (string)recycleTime.Value;
					var timeParts = value.Split(':');
					var time = DateTime.Today.AddHours(int.Parse(timeParts[0])).AddMinutes(int.Parse(timeParts[1]));
					// increase time by one minute, so the warmup occures after recycling
					time = time.AddMinutes(1);
					return time;
				}
			}
			catch (Exception ex)
			{

				return DateTime.Today.AddHours(3);
			}
		}

		private void AddJavaScript(string javaScriptFileName, string sodKey, string sodKeyDependency)
		{
			var link = "/_admin/RH.Warmup/" + javaScriptFileName;
			if (string.IsNullOrEmpty(sodKey))
			{
				ScriptManager.RegisterClientScriptInclude(this, typeof(TimePicker), javaScriptFileName, link);
				return;
			}

			var sb = new StringBuilder();
			sb.AppendFormat("RegisterSod('{0}', '{1}');{2}", sodKey, link, Environment.NewLine);
			if (!string.IsNullOrEmpty(sodKeyDependency))
			{
				sb.AppendFormat("RegisterSodDep('{0}', '{1}');{2}", sodKey, sodKeyDependency, Environment.NewLine);
			}
			
			// to have the script loaded, need to ensure it
			//sb.AppendFormat("EnsureScriptFunc('{0}');{1}", sodKey, Environment.NewLine);

			ScriptManager.RegisterStartupScript(this, typeof(TimePicker), javaScriptFileName + sodKey + sodKeyDependency, sb.ToString(), true);
		}
	}
}
