﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Net;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using WarmupSharePoint.Helper;

namespace WarmupSharePoint
{
	public class Authentication
	{
		private Authentication(string absoluteUrl)
		{
			AbsoluteUrl = absoluteUrl;
		}

		public string AbsoluteUrl { get; private set; }
		public bool IsClaims { get; private set; }
		public bool IsWindowsAvailable { get; private set; }
		public ICredentials Credentials { get; private set; }

		public static Authentication GetFromUrl(string absoluteUrl)
		{
			if (string.IsNullOrEmpty(absoluteUrl)) throw new ArgumentNullException("absoluteUrl");
			var authentication = new Authentication(absoluteUrl);

			Uri requestUri;
			if (!Uri.TryCreate(absoluteUrl, UriKind.Absolute, out requestUri))
				throw new ArgumentException(absoluteUrl + " is no a valid URL.");

			using (var site = new SPSite(requestUri.AbsoluteUri))
			{
				SPUrlZone urlZone;
				SPWebApplication webApplication = site.WebApplication;
				try
				{
					urlZone = webApplication.AlternateUrls[absoluteUrl].UrlZone;
				}
				catch (Exception ex)
				{
					Trace.WriteLine("Cannot find an AAM for " + absoluteUrl);
					string[] availableUrls = webApplication.AlternateUrls.Select(au => au.Uri.AbsoluteUri).ToArray();
					throw new ArgumentException("The URL '" + absoluteUrl + "' is not a valid AAM. The configured AAMs are " + string.Join(",", availableUrls), ex);
				}

				SPIisSettings iisSettings;
				try
				{
					iisSettings = webApplication.IisSettings[urlZone];
				}
				catch (KeyNotFoundException ex)
				{
					string[] iisUrls = SharePointHelper.GetAlternatUrlForWebApplication(webApplication);

					var availableAaMs = string.Join(",", iisUrls);
					if (iisUrls.Length > 1)
						throw new ArgumentException("Please use one of the configured URLs in IIS, as all other URLs are aliases: " + availableAaMs);

					// if only one url is available, the webapplication has not been extenden. use the first url found
					urlZone = webApplication.AlternateUrls[iisUrls[0]].UrlZone;
					iisSettings = webApplication.IisSettings[urlZone];
				}

				authentication.IsWindowsAvailable = iisSettings.UseWindowsIntegratedAuthentication;

				// needed for web request later
				authentication.IsClaims = webApplication.UseClaimsAuthentication;

				return authentication;
			}
		}
		public static Authentication GetFromUrl(string absoluteUrl, KeyValuePair<SPUrlZone, SPIisSettings> iisSetting, SPWebApplication webApplication)
		{
			var authentication = new Authentication(absoluteUrl)
			{
				IsWindowsAvailable = iisSetting.Value.UseWindowsIntegratedAuthentication,
				// needed for web request later
				IsClaims = webApplication.UseClaimsAuthentication
			};

			return authentication;
		}

		internal ICredentials GetCredentials()
		{
			return CredentialCache.DefaultCredentials;
		}

	}
}