using System;
using System.Diagnostics;

namespace WarmupSharePoint
{
	class Program
	{
		private static readonly Version Version = new Version(2, 0, 0, 0);

		private static string SearchKeyword
		{
			get
			{
				if (Properties.Settings.Default != null)
					return Properties.Settings.Default.SearchKeyword;

				return "WarmupSharePoint";
			}
		}

		static void Main(string[] args)
		{
			try
			{
				Console.WriteLine("WarmupSharePoint ({0}) by Ren Hzser", Version);
				WarmupScope warmupScope = GetScopeFromParameters(args);
				if (warmupScope == WarmupScope.Invalid)
				{
					Usage.Show();
					return;
				}

				var appSettings = Properties.Settings.Default;
				var settings = new Settings { EventLog = appSettings.Eventlog, WarmupSearch = appSettings.EnableSearchWarmup, QueryKeyword = SearchKeyword };
				if (args.Length > 0 && warmupScope != WarmupScope.Farm)
					settings.WebApplicationUrl = args[0];

				Eventlog.Write(string.Format("WarmupSharePoint ({0}) by Ren Hzser started with Scope \"{1}\" {2}", Version, warmupScope, args.Length > 0 ? "and URL " + args[0] : string.Empty));
				var stopwatch = new Stopwatch();
				stopwatch.Start();
				var warmup = new Warmup(settings);
				switch (warmupScope)
				{
					case WarmupScope.Farm:
						warmup.WarmupFarm();
						break;
					case WarmupScope.SingleSiteCollection:
						Authentication authenticationInfos = Authentication.GetFromUrl(args[0]);
						warmup.WarmupWebsFromSite(args[0], authenticationInfos);
						break;
					case WarmupScope.WebApplication:
						Console.WriteLine("Starting with {0}...", args[0]);
						warmup.ReadAllSites();
						break;
				}
				stopwatch.Stop();
				var timespan = stopwatch.Elapsed;
				var duration = string.Format("{3} web(s) in {0:00}h:{1:00}m:{2:00}s", timespan.Hours, timespan.Minutes, timespan.Seconds, warmup.WarmupWebsCount);
				Console.WriteLine(duration);
				EventLogEntryType type = warmup.WarmupErrors.Length == 0 ? EventLogEntryType.Information : EventLogEntryType.Error;
				Eventlog.Write(string.Format("WarmupSharePoint ({0}) by Ren Hzser finished.{2}{1}{2}{2}{3}",
					Version, duration, Environment.NewLine, warmup.WarmupErrors), type, warmup.WarmupErrors.Length == 0 ? 0 : 1);
			}
			catch (Exception ex)
			{
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Error in WarmupSharePoint:\r\n{0}", ex);
				Console.ForegroundColor = ConsoleColor.Gray;
				Eventlog.Write("WarmupSharePoint has a problem: " + ex);
			}
		}

		/// <summary>
		/// identify WarmupScope from parameters
		/// </summary>
		/// <param name="args"></param>
		/// <returns></returns>
		private static WarmupScope GetScopeFromParameters(string[] args)
		{
			var warmupScope = WarmupScope.Invalid;
			var argumentCount = args.Length;

			if (argumentCount == 1 && args[0].Equals("Farm", StringComparison.InvariantCultureIgnoreCase))
			{
				warmupScope = WarmupScope.Farm;
			}
			else if (argumentCount == 1)
			{
				warmupScope = WarmupScope.SingleSiteCollection;
			}
			else if (argumentCount == 2 && args[1].ToUpper() == "ALLSITES")
			{
				warmupScope = WarmupScope.WebApplication;
			}

			return warmupScope;
		}
	}
}