﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.SharePoint.Utilities;
using WarmupSharePoint.Helper;

namespace WarmupSharePoint
{
	public class Warmup
	{
		private readonly Settings _settings;
		public event EventHandler<ProgressChangeEventArgs> ProgressChanged;

		internal Warmup()
		{
			WarmupWebsCount = 0;
			WarmupErrors = new StringBuilder();
		}

		public Warmup(Settings settings)
			: this()
		{
			_settings = settings;
		}

		public int WarmupWebsCount;
		public int AllWebsCount;
		public StringBuilder WarmupErrors;

		protected virtual void OnProgressChanged(ProgressChangeEventArgs e)
		{
			if (ProgressChanged != null)
				ProgressChanged(this, e);
		}

		/// <summary>
		/// Warmup all SiteCollections of the farm in all IISSettings
		/// </summary>
		public void WarmupFarm()
		{
			using (new SPMonitoredScope("WarmupSharePoint.Farm.WarmupFarm()"))
			{
				Console.WriteLine("Working on the whole farm...");
				SPWebService contentService = SPWebService.ContentService;
				foreach (SPWebApplication webApplication in contentService.WebApplications)
				{
					AllWebsCount += SharePointHelper.GetWebsCountFromWebApplication(webApplication);
					foreach (KeyValuePair<SPUrlZone, SPIisSettings> iisSetting in webApplication.IisSettings)
					{
						SPAlternateUrl incomingUrl = webApplication.AlternateUrls.GetResponseUrl(iisSetting.Key);

						var authenticationInfos = Authentication.GetFromUrl(incomingUrl.IncomingUrl, iisSetting, webApplication);
						foreach (SPSite site in webApplication.Sites)
						{
							try
							{
								string incomingSiteUrl = webApplication.Farm.AlternateUrlCollections.RebaseUriWithAlternateUri(new Uri(site.Url), iisSetting.Key).AbsoluteUri.TrimEnd('/');
								Console.WriteLine("Site {0}...", incomingSiteUrl);
								WarmupWebsFromSite(incomingSiteUrl, authenticationInfos);
							}
							catch (Exception ex)
							{
								Console.WriteLine(ex.Message);
								WarmupErrors.AppendFormat("{0} - {1}", site.Url, ex.Message);
							}
							finally
							{
								site.Dispose();
							}
						}
					}
				}
			}
		}

		/// <summary>
		/// Warmup all SPWebs within the passed SPSiteUrl
		/// </summary>
		/// <param name="rootUrl">used for creating a new SPSite object</param>
		/// <param name="authenticationInfos"></param>
		public void WarmupWebsFromSite(string rootUrl, Authentication authenticationInfos)
		{
			if (string.IsNullOrEmpty(rootUrl)) throw new ArgumentNullException("rootUrl");
			if (authenticationInfos == null) throw new ArgumentNullException("authenticationInfos");

			using (new SPMonitoredScope("WarmupSharePoint.Farm.WarmupWebsFromSite(" + rootUrl + ")"))
			{
				try
				{
					using (var site = new SPSite(rootUrl))
					using (SPWeb rootWeb = site.RootWeb)
					{
						Console.Write("  {0} ", rootWeb.Title);

						if (AllWebsCount == 0)
						{
							// set overall web count for running this method only. in other cases, the property is filled elsethere
							AllWebsCount = site.AllWebs.Count;
						}

						ConsoleColor color;
						var warmupResult = CallWebsite(site, rootWeb.Url, authenticationInfos, false, out color);
						Console.ForegroundColor = color;
						Console.Write(warmupResult + Environment.NewLine);
						Console.ForegroundColor = ConsoleColor.Gray;

						SearchWarmup(site, rootWeb, authenticationInfos);

						foreach (SPWeb web in rootWeb.GetSubwebsForCurrentUser())
						{
							try
							{
								WarmupWeb(site, web, authenticationInfos);
								SearchWarmup(site, web, authenticationInfos);

								ReadSubWebsForCurrentUser(site, web, authenticationInfos);
							}
							finally
							{
								web.Dispose();
							}
						}
					}
				}
				catch (Exception ex)
				{
					Console.ForegroundColor = ConsoleColor.Red;
					Console.WriteLine(ex.Message);
					Console.ForegroundColor = ConsoleColor.Gray;
					WarmupErrors.AppendFormat("{0} - {1}", rootUrl, ex.Message);
				}
			}
		}

		/// <summary>
		/// Warmup all Subwebs for the current user
		/// </summary>
		/// <param name="site"></param>
		/// <param name="web"></param>
		/// <param name="authenticationInfos"></param>
		private void ReadSubWebsForCurrentUser(SPSite site, SPWeb web, Authentication authenticationInfos)
		{
			foreach (SPWeb subWeb in web.GetSubwebsForCurrentUser())
			{
				try
				{
					WarmupWeb(site, subWeb, authenticationInfos);
					ReadSubWebsForCurrentUser(site, subWeb, authenticationInfos);
				}
				finally
				{
					subWeb.Dispose();
				}
			}
		}

		/// <summary>
		/// Warmup all SiteCollections of the passed Url of a WebApplication
		/// </summary>
		public void ReadAllSites()
		{
			using (new SPMonitoredScope("WarmupSharePoint.Farm.ReadAllSites(" + _settings.WebApplicationUrl + ")"))
			{
				using (var outerSite = new SPSite(_settings.WebApplicationUrl))
				{
					var webApplication = outerSite.WebApplication;

					AllWebsCount = SharePointHelper.GetWebsCountFromWebApplication(webApplication);
					WarmupWebsCount = 0;
					var authenticationInfos = Authentication.GetFromUrl(_settings.WebApplicationUrl);
					SPSiteCollection sites = webApplication.Sites;
					foreach (SPSite subSite in sites)
					{
						try
						{
							using (SPWeb rootWeb = subSite.RootWeb)
							{
								Console.WriteLine("Working on '{0}'", rootWeb.Url);
								WarmupWebsFromSite(rootWeb.Url, authenticationInfos);
							}
						}
						catch (Exception ex)
						{
							Console.WriteLine(ex.Message);
							WarmupErrors.AppendFormat("{0} - {1}", subSite.Url, ex.Message);
						}
						finally
						{
							subSite.Dispose();
						}
					}
				}
			}
		}

		/// <summary>
		/// acutally warms up the web
		/// </summary>
		/// <param name="site"></param>
		/// <param name="web"></param>
		/// <param name="authenticationInfos"></param>
		private void WarmupWeb(SPSite site, SPWeb web, Authentication authenticationInfos)
		{
			// count ParentWebs of the web, to indent console output
			var depth = web.ServerRelativeUrl.Count(c => c.Equals('/'));
			string indent = string.Empty;
			for (int i = 0; i < depth; i++)
			{
				indent += "  ";
			}
			Console.Write("{0}{1} ", indent, web.Title);
			ConsoleColor color;
			var warmupResult = CallWebsite(site, web.Url, authenticationInfos, false, out color);
			Console.ForegroundColor = color;
			Console.Write(warmupResult + Environment.NewLine);
			Console.ForegroundColor = ConsoleColor.Gray;
		}

		/// <summary>
		/// warmup SearchCenter (if the web is one) and the Foundation Search (if web is RootWeb)
		/// </summary>
		/// <param name="site"></param>
		/// <param name="web"></param>
		/// <param name="authenticationInfos"></param>
		private void SearchWarmup(SPSite site, SPWeb web, Authentication authenticationInfos)
		{
			if (Properties.Settings.Default != null && Properties.Settings.Default.EnableSearchWarmup == false) return;

			ConsoleColor color;
			// warmup SearchCenter
			string warmupResult = WarmupSearchCenter(site, web, authenticationInfos, out color);
			if (!string.IsNullOrEmpty(warmupResult))
			{
				Console.Write("  {0} (SearchCenter)...", web.Title);
				Console.ForegroundColor = color;
				Console.Write(warmupResult + Environment.NewLine);
				Console.ForegroundColor = ConsoleColor.Gray;
			}

			if (!web.IsRootWeb) return;

			// warmup Foundation Search only necessary on RootWeb
			var searchUrl = SPUrlUtility.CombineUrl(web.Url, "/_layouts/OSSSearchResults.aspx?k=" + _settings.QueryKeyword);
			Console.Write("  Foundation Search will be warmed up ");
			warmupResult = CallWebsite(site, searchUrl, authenticationInfos, true, out color);
			Console.ForegroundColor = color;
			Console.Write(warmupResult + Environment.NewLine);
			Console.ForegroundColor = ConsoleColor.Gray;
		}

		/// <summary>
		/// Warmup SearchCenter by calling the SearchResult page with a keyword
		/// </summary>
		/// <param name="site"></param>
		/// <param name="web"></param>
		/// <param name="authenticationInfos"></param>
		/// <param name="color"></param>
		private string WarmupSearchCenter(SPSite site, SPWeb web, Authentication authenticationInfos, out ConsoleColor color)
		{
			string result = string.Empty;
			color = ConsoleColor.Gray;

			// check for search center or fast search center
			if (web.WebTemplateId == 50 || web.WebTemplateId == 90 || // SharePoint Search Centers
				web.WebTemplateId == 2000) // FAST SearchCenter
			{
				var documentLibraries = web.GetListsOfType(SPBaseType.DocumentLibrary).Cast<SPList>();
				var list = documentLibraries.FirstOrDefault(l => l.BaseTemplate == (SPListTemplateType)850);
				if (list == null)
				{
					// a basic searchcenter has its results page in the root directory
					foreach (SPFile file in web.RootFolder.Files)
					{
						if (file.Name.Equals("results.aspx", StringComparison.InvariantCultureIgnoreCase))
						{
							// search result page found
							var searchResultUrl = web.Site.MakeFullUrl(file.ServerRelativeUrl) + "?k=" + _settings.QueryKeyword;
							Trace.WriteLine("Found SearchResultPage '" + searchResultUrl + "'");
							result = CallWebsite(site, searchResultUrl, authenticationInfos, true, out color);
							break;
						}
					}
					return result;
				}

				foreach (SPListItem item in list.Items)
				{
					result += CheckSearchLayoutAndWarmup(site, web, item, authenticationInfos, out color);
					if (color != ConsoleColor.Black) break;
				}
			}

			return result;
		}

		/// <summary>
		/// if the item uses 'Search results' as pagelayout, warmup the search
		/// </summary>
		/// <param name="site"></param>
		/// <param name="web"></param>
		/// <param name="item"></param>
		/// <param name="authenticationInfos"></param>
		/// <param name="color">wll be black, if not results page was found</param>
		/// <returns></returns>
		private string CheckSearchLayoutAndWarmup(SPSite site, SPWeb web, SPListItem item, Authentication authenticationInfos, out ConsoleColor color)
		{
			string result = string.Empty;
			color = ConsoleColor.Black;

			var layout = Convert.ToString(item["PublishingPageLayout"]);
			if (layout.EndsWith("Search results", StringComparison.InvariantCultureIgnoreCase))
			{
				// search result page found
				var searchResultUrl = web.Site.MakeFullUrl(item.File.ServerRelativeUrl) + "?k=" + _settings.QueryKeyword;
				Trace.WriteLine("Found SearchResultPage '" + searchResultUrl + "'");
				result = CallWebsite(site, searchResultUrl, authenticationInfos, true, out color);
			}
			return result;
		}

		/// <summary>
		/// calls the Website by its URL and reads the Response to "warm it up"
		/// </summary>
		/// <param name="site">Site is used to get AuthenticationProvider information from it</param>
		/// <param name="url">has to be an absolute url</param>
		/// <param name="authenticationInfos"></param>
		/// <param name="isSearch"></param>
		/// <param name="color"> </param>
		/// <returns>OK | ex.Message</returns>
		public string CallWebsite(SPSite site, string url, Authentication authenticationInfos, bool isSearch, out ConsoleColor color)
		{
			if (authenticationInfos == null) throw new ArgumentNullException("authenticationInfos");
			try
			{
				MakeRequest(site, url, authenticationInfos);

				if (!isSearch)
				{
					WarmupWebsCount++;
					if (WarmupWebsCount > AllWebsCount)
						WarmupWebsCount = AllWebsCount;

					OnProgressChanged(new ProgressChangeEventArgs(AllWebsCount, WarmupWebsCount));
				}
				color = ConsoleColor.DarkGreen;
				return "OK";
			}
			catch (WebException ex)
			{
				var message = WebHelper.GetWebExceptionMessage(ex);
				WarmupErrors.AppendFormat("{0} - {1}{2}", url, message, Environment.NewLine);
				color = ConsoleColor.DarkRed;
				return ex.Message;
			}
			catch (Exception ex)
			{
				WarmupErrors.AppendFormat("{0} - {1}{2}", url, ex.Message, Environment.NewLine);
				color = ConsoleColor.DarkRed;
				return ex.Message;
			}
		}

		private static void MakeRequest(SPSite site, string url, Authentication authenticationInfos)
		{
			if (!authenticationInfos.IsClaims)
			{
				using (var client = new WebClient())
				{
					client.Credentials = CredentialCache.DefaultNetworkCredentials;
					client.DownloadData(url);
				}
			}
			else
			{
				try
				{
					MakeClaimsRequest(url, authenticationInfos);
				}
				catch (WebException ex)
				{
					if (ex.Status == WebExceptionStatus.ProtocolError)
					{
						// e.g. Server Unavailable
						throw;
					}

					var logonUrl = GetLoginPageUrl(site, url);
					MakeClaimsRequest(logonUrl, authenticationInfos);
				}
			}
		}

		private static void MakeClaimsRequest(string url, Authentication authenticationInfos)
		{
			//if we're in claims mode, we have to utilize a HttpWebRequest object to properly pass user credentials to the form generation page 
			var request = (HttpWebRequest)WebRequest.Create(url);
			request.UserAgent= "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)";
			request.PreAuthenticate = true;
			request.CookieContainer = new CookieContainer();
			request.Credentials = authenticationInfos.GetCredentials();
			// allows redirecting from loginpage to requested page
			request.AllowAutoRedirect = true;

			NameValueCollection headers = request.Headers;
			//we have to make sure to pass our current WIF authorization cookie to the request, otherwise we get a 403 error 
			headers.Add("Cookie", headers["Cookie"]);
			// use FBA
			headers.Add("X-FORMS_BASED_AUTH_ACCEPTED", "f");

			//allows for self-signed SSL certificates on the response if our configuration is set properly 
			ServicePointManager.ServerCertificateValidationCallback += ValidateRemoteCertificate;

			// authentication request
			var cookieContainer = new CookieCollection();
			using (var response = request.GetResponse() as HttpWebResponse)
			{
				if (response != null && request.HaveResponse)
				{
					cookieContainer = response.Cookies;
					if (response.StatusCode != HttpStatusCode.Moved)
					{
						// we may warmup 
						using (Stream responseStream = response.GetResponseStream())
						using (var streamReader = new StreamReader(responseStream))
						{
							string html = streamReader.ReadToEnd();
						}
					}
				}
			}
		}

		private static string GetLoginPageUrl(SPSite site, string url)
		{
			SPUrlZone urlZone = site.WebApplication.AlternateUrls[url].UrlZone;
			var iisSettings = site.WebApplication.IisSettings[urlZone];
			if (iisSettings.UseWindowsIntegratedAuthentication)
			{
				// maybe use later. for now, use fixed url for performance reason
				//SPIisSettings settings = site.WebApplication.IisSettings[urlZone];
				//SPAuthenticationProvider windowsAuthProvider = (from provider in settings.ClaimsAuthenticationProviders
				//																let windowsAuthenticationProvider = provider as SPWindowsAuthenticationProvider
				//																where windowsAuthenticationProvider != null && windowsAuthenticationProvider.UseWindowsIntegratedAuthentication
				//																select provider).First();
				var authenticationProviderUrl = "_windows/default.aspx"; //windowsAuthProvider.AuthenticationRedirectionUrl.ToString()

				// create logon url with redirect to original site url
				var encodedSiteUrl = SPEncode.UrlEncode(site.ServerRelativeUrl);
				var logonUrl = string.Format("{0}?ReturnUrl=%2f_layouts%2fAuthenticate.aspx%3fSource%3d{2}&Source={1}",
					SPUrlUtility.CombineUrl(site.Url, authenticationProviderUrl), encodedSiteUrl, SPEncode.UrlEncode(encodedSiteUrl));
				return logonUrl;
			}

			if (iisSettings.AllowAnonymous)
				return url;

			throw new NotSupportedException("The site " + url + " does not use Windows Authentication. Cannot login with integrated credentials.");
		}

		/// <summary>
		/// callback used to validate the certificate in an SSL conversation 
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="certificate"></param>
		/// <param name="chain"></param>
		/// <param name="policyErrors"></param>
		/// <returns></returns>
		private static bool ValidateRemoteCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors policyErrors)
		{
			// allow any certificate return true;
			return true;
		}
	}
}
